@@ -6,19 +6,24 @@ from typing import Any, Dict, Iterable, Optional, Union
 
 from ..client import Client, register_client_class
 from ..cloudpath import implementation_registry
+from ..exceptions import MissingCredentialsError
 from .azblobpath import AzureBlobPath
 
 
 try:
     from azure.core.exceptions import ResourceNotFoundError
-    from azure.storage.blob import BlobServiceClient
+    from azure.storage.blob import BlobServiceClient, BlobProperties
 except ModuleNotFoundError:
     implementation_registry["azure"].dependencies_loaded = False
 
 
 @register_client_class("azure")
 class AzureBlobClient(Client):
-    """Client for Azure Blob Storage."""
+    """Client class for Azure Blob Storage which handles authentication with Azure for
+    [`AzureBlobPath`](../azblobpath/) instances. See documentation for the
+    [`__init__` method][cloudpathlib.azure.azblobclient.AzureBlobClient.__init__] for detailed
+    authentication options.
+    """
 
     def __init__(
         self,
@@ -43,7 +48,8 @@ class AzureBlobClient(Client):
         https://docs.microsoft.com/en-us/python/api/azure-storage-blob/azure.storage.blob.blobserviceclient?view=azure-python).
 
         If multiple methods are used, priority order is reverse of list above (later in list takes
-        priority).
+        priority). If no methods are used, a [`MissingCredentialsError`][cloudpathlib.exceptions.MissingCredentialsError]
+        exception will be raised raised.
 
         Args:
             account_url (Optional[str]): The URL to the blob storage account, optionally
@@ -63,6 +69,8 @@ class AzureBlobClient(Client):
             local_cache_dir (Optional[Union[str, os.PathLike]]): Path to directory to use as cache
                 for downloaded files. If None, will use a temporary directory.
         """
+        if connection_string is None:
+            connection_string = os.getenv("AZURE_STORAGE_CONNECTION_STRING", None)
 
         if blob_service_client is not None:
             self.service_client = blob_service_client
@@ -73,13 +81,14 @@ class AzureBlobClient(Client):
         elif account_url is not None:
             self.service_client = BlobServiceClient(account_url=account_url, credential=credential)
         else:
-            self.service_client = BlobServiceClient.from_connection_string(
-                os.getenv("AZURE_STORAGE_CONNECTION_STRING")
+            raise MissingCredentialsError(
+                "AzureBlobClient does not support anonymous instantiation. "
+                "Credentials are required; see docs for options."
             )
 
         super().__init__(local_cache_dir=local_cache_dir)
 
-    def _get_metadata(self, cloud_path: AzureBlobPath) -> Dict[str, Any]:
+    def _get_metadata(self, cloud_path: AzureBlobPath) -> Union[BlobProperties, Dict[str, Any]]:
         blob = self.service_client.get_blob_client(
             container=cloud_path.container, blob=cloud_path.blob
         )
@@ -89,7 +98,7 @@ class AzureBlobClient(Client):
 
     def _download_file(
         self, cloud_path: AzureBlobPath, local_path: Union[str, os.PathLike]
-    ) -> Union[str, os.PathLike]:
+    ) -> Path:
         blob = self.service_client.get_blob_client(
             container=cloud_path.container, blob=cloud_path.blob
         )
@@ -162,7 +171,9 @@ class AzureBlobClient(Client):
 
             yield self.CloudPath(f"az://{cloud_path.container}/{o.name}")
 
-    def _move_file(self, src: AzureBlobPath, dst: AzureBlobPath) -> AzureBlobPath:
+    def _move_file(
+        self, src: AzureBlobPath, dst: AzureBlobPath, remove_src: bool = True
+    ) -> AzureBlobPath:
         # just a touch, so "REPLACE" metadata
         if src == dst:
             blob_client = self.service_client.get_blob_client(
@@ -180,13 +191,14 @@ class AzureBlobClient(Client):
 
             target.start_copy_from_url(source.url)
 
-            self._remove(src)
+            if remove_src:
+                self._remove(src)
 
         return dst
 
     def _remove(self, cloud_path: AzureBlobPath) -> None:
         if self._is_file_or_dir(cloud_path) == "dir":
-            blobs = [b.blob for b in self._list_dir(cloud_path, recursive=True)]
+            blobs = [b.blob for b in self._list_dir(cloud_path, recursive=True) if b.is_file()]
             container_client = self.service_client.get_container_client(cloud_path.container)
             container_client.delete_blobs(*blobs)
         elif self._is_file_or_dir(cloud_path) == "file":
