@@ -8,53 +8,28 @@ from typing import Any, IO, Iterable, Optional, TYPE_CHECKING, Union
 from urllib.parse import urlparse
 from warnings import warn
 
+from . import anypath
+
+from .exceptions import (
+    ClientMismatchError,
+    CloudPathFileExistsError,
+    CloudPathIsADirectoryError,
+    CloudPathNotADirectoryError,
+    DirectoryNotEmptyError,
+    IncompleteImplementationError,
+    InvalidPrefixError,
+    MissingDependenciesError,
+    NoStatError,
+    OverwriteDirtyFileError,
+    OverwriteNewerCloudError,
+    OverwriteNewerLocalError,
+)
+
 
 if TYPE_CHECKING:
     from .client import Client
 
 
-# Custom Exceptions
-class MissingDependencies(Exception):
-    pass
-
-
-class IncompleteImplementation(NotImplementedError):
-    pass
-
-
-class ClientMismatch(ValueError):
-    pass
-
-
-class InvalidPrefix(ValueError):
-    pass
-
-
-class OverwriteDirtyFile(Exception):
-    pass
-
-
-class OverwriteNewerCloud(Exception):
-    pass
-
-
-class OverwriteNewerLocal(Exception):
-    pass
-
-
-class DirectoryNotEmpty(Exception):
-    pass
-
-
-class NoStat(Exception):
-    """Used if stats cannot be retrieved; e.g., file does not exist
-    or for some backends path is a directory (which doesn't have
-    stats available).
-    """
-
-    pass
-
-
 class CloudImplementation:
     def __init__(self):
         self.name = None
@@ -66,11 +41,11 @@ class CloudImplementation:
         expected = ["client_class", "path_class"]
         missing = [cls for cls in expected if getattr(self, f"_{cls}") is None]
         if missing:
-            raise IncompleteImplementation(
+            raise IncompleteImplementationError(
                 f"Implementation is missing registered components: {missing}"
             )
         if not self.dependencies_loaded:
-            raise MissingDependencies(
+            raise MissingDependenciesError(
                 f"Missing dependencies for {self._client_class.__name__}. You can install them "
                 f"with 'pip install cloudpathlib[{self.name}]'."
             )
@@ -122,7 +97,7 @@ class CloudPathMeta(abc.ABCMeta):
                 for impl in implementation_registry.values()
                 if impl._path_class is not None
             ]
-            raise InvalidPrefix(
+            raise InvalidPrefixError(
                 f"Path {cloud_path} does not begin with a known prefix " f"{valid}."
             )
 
@@ -150,6 +125,20 @@ class CloudPathMeta(abc.ABCMeta):
 
 # Abstract base class
 class CloudPath(metaclass=CloudPathMeta):
+    """Base class for cloud storage file URIs, in the style of the Python standard library's
+    [`pathlib` module](https://docs.python.org/3/library/pathlib.html). Instances represent a path
+    in cloud storage with filesystem path semantics, and convenient methods allow for basic
+    operations like joining, reading, writing, iterating over contents, etc. `CloudPath` almost
+    entirely mimics the [`pathlib.Path`](https://docs.python.org/3/library/pathlib.html#pathlib.Path)
+    interface, so most familiar properties and methods should be available and behave in the
+    expected way.
+
+    Analogous to the way `pathlib.Path` works, instantiating `CloudPath` will instead create an
+    instance of an appropriate subclass that implements a particular cloud storage service, such as
+    [`S3Path`](../s3path). This dispatching behavior is based on the URI scheme part of a cloud
+    storage URI (e.g., `"s3://"`).
+    """
+
     _cloud_meta: CloudImplementation
     cloud_prefix: str
 
@@ -168,7 +157,7 @@ class CloudPath(metaclass=CloudPathMeta):
             else:
                 client = self._cloud_meta.client_class.get_default_client()
         if not isinstance(client, self._cloud_meta.client_class):
-            raise ClientMismatch(
+            raise ClientMismatchError(
                 f"Client of type [{client.__class__}] is not valid for cloud path of type "
                 f"[{self.__class__}]; must be instance of [{self._cloud_meta.client_class}], or "
                 f"None to use default client for this cloud path class."
@@ -199,30 +188,49 @@ class CloudPath(metaclass=CloudPathMeta):
         valid = str(path).lower().startswith(cls.cloud_prefix.lower())
 
         if raise_on_error and not valid:
-            raise InvalidPrefix(
+            raise InvalidPrefixError(
                 f"'{path}' is not a valid path since it does not start with '{cls.cloud_prefix}'"
             )
 
         return valid
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f"{self.__class__.__name__}('{self}')"
 
-    def __str__(self):
+    def __str__(self) -> str:
         return self._str
 
-    def __hash__(self):
-        # use repr for type and resolved path to assess if paths are the same
-        return hash(self.__repr__)
+    def __hash__(self) -> int:
+        return hash((type(self).__name__, str(self)))
 
-    def __eq__(self, other: Any):
-        return repr(self) == repr(other)
+    def __eq__(self, other: Any) -> bool:
+        return isinstance(other, type(self)) and str(self) == str(other)
 
     def __fspath__(self):
         if self.is_file():
             self._refresh_cache(force_overwrite_from_cloud=False)
         return str(self._local)
 
+    def __lt__(self, other: Any) -> bool:
+        if not isinstance(other, type(self)):
+            return NotImplemented
+        return self.parts < other.parts
+
+    def __le__(self, other: Any) -> bool:
+        if not isinstance(other, type(self)):
+            return NotImplemented
+        return self.parts <= other.parts
+
+    def __gt__(self, other: Any) -> bool:
+        if not isinstance(other, type(self)):
+            return NotImplemented
+        return self.parts > other.parts
+
+    def __ge__(self, other: Any) -> bool:
+        if not isinstance(other, type(self)):
+            return NotImplemented
+        return self.parts >= other.parts
+
     # ====================== NOT IMPLEMENTED ======================
     # absolute - no cloud equivalent; all cloud paths are absolute already
     # as_posix - no cloud equivalent; not needed since we assume url separator
@@ -332,10 +340,12 @@ class CloudPath(metaclass=CloudPathMeta):
     ) -> IO:
         # if trying to call open on a directory that exists
         if self.exists() and not self.is_file():
-            raise ValueError(f"Cannot open directory, only files. Tried to open ({self})")
+            raise CloudPathIsADirectoryError(
+                f"Cannot open directory, only files. Tried to open ({self})"
+            )
 
         if mode == "x" and self.exists():
-            raise ValueError(f"Cannot open existing file ({self}) for creation.")
+            raise CloudPathFileExistsError(f"Cannot open existing file ({self}) for creation.")
 
         # TODO: consider streaming from client rather than DLing entire file to cache
         self._refresh_cache(force_overwrite_from_cloud=force_overwrite_from_cloud)
@@ -386,7 +396,7 @@ class CloudPath(metaclass=CloudPathMeta):
 
     def replace(self, target: "CloudPath") -> "CloudPath":
         if type(self) != type(target):
-            raise ValueError(
+            raise TypeError(
                 f"The target based to rename must be an instantiated class of type: {type(self)}"
             )
 
@@ -406,10 +416,12 @@ class CloudPath(metaclass=CloudPathMeta):
 
     def rmdir(self):
         if self.is_file():
-            raise ValueError(f"Path {self} is a file; call unlink instead of rmdir.")
+            raise CloudPathNotADirectoryError(
+                f"Path {self} is a file; call unlink instead of rmdir."
+            )
         try:
             next(self.iterdir())
-            raise DirectoryNotEmpty(
+            raise DirectoryNotEmptyError(
                 f"Directory not empty: '{self}'. Use rmtree to delete recursively."
             )
         except StopIteration:
@@ -422,7 +434,9 @@ class CloudPath(metaclass=CloudPathMeta):
 
     def unlink(self):
         if self.is_dir():
-            raise ValueError(f"Path {self} is a directory; call rmdir instead of unlink.")
+            raise CloudPathIsADirectoryError(
+                f"Path {self} is a directory; call rmdir instead of unlink."
+            )
         self.client._remove(self)
 
     def write_bytes(self, data: bytes):
@@ -573,10 +587,12 @@ class CloudPath(metaclass=CloudPathMeta):
         return self._dispatch_to_local_cache_path("read_text")
 
     # ===========  public cloud methods, not in pathlib ===============
-    def download_to(self, destination: Union[str, os.PathLike]):
+    def download_to(self, destination: Union[str, os.PathLike]) -> Path:
         destination = Path(destination)
         if self.is_file():
-            self.client._download_file(self, destination)
+            if destination.is_dir():
+                destination = destination / self.name
+            return self.client._download_file(self, destination)
         else:
             destination.mkdir(exist_ok=True)
             for f in self.iterdir():
@@ -587,12 +603,118 @@ class CloudPath(metaclass=CloudPathMeta):
                 rel_dest = str(f)[len(rel) :]
                 f.download_to(destination / rel_dest)
 
+            return destination
+
     def rmtree(self):
         """Delete an entire directory tree."""
         if self.is_file():
-            raise ValueError(f"Path {self} is a file; call unlink instead of rmtree.")
+            raise CloudPathNotADirectoryError(
+                f"Path {self} is a file; call unlink instead of rmtree."
+            )
         self.client._remove(self)
 
+    def upload_from(
+        self, source: Union[str, os.PathLike], force_overwrite_to_cloud: bool = False
+    ) -> "CloudPath":
+        """Upload a file or directory to the cloud path."""
+        source = Path(source)
+
+        if source.is_dir():
+            for p in source.iterdir():
+                (self / p.name).upload_from(p, force_overwrite_to_cloud=force_overwrite_to_cloud)
+
+            return self
+
+        else:
+            if self.exists() and self.is_dir():
+                dst = self / source.name
+            else:
+                dst = self
+
+            dst._upload_file_to_cloud(source, force_overwrite_to_cloud=force_overwrite_to_cloud)
+
+            return dst
+
+    def copy(
+        self,
+        destination: Union[str, os.PathLike, "CloudPath"],
+        force_overwrite_to_cloud: bool = False,
+    ) -> Union[Path, "CloudPath"]:
+        """Copy self to destination folder of file, if self is a file."""
+        if not self.exists() or not self.is_file():
+            raise ValueError(
+                f"Path {self} should be a file. To copy a directory tree use the method copytree."
+            )
+
+        # handle string version of cloud paths + local paths
+        if isinstance(destination, (str, os.PathLike)):
+            destination = anypath.to_anypath(destination)
+
+        if not isinstance(destination, CloudPath):
+            return self.download_to(destination)
+
+        # if same client, use cloud-native _move_file on client to avoid downloading
+        elif self.client is destination.client:
+            if destination.exists() and destination.is_dir():
+                destination: CloudPath = destination / self.name  # type: ignore
+
+            if (
+                not force_overwrite_to_cloud
+                and destination.exists()
+                and destination.stat().st_mtime >= self.stat().st_mtime
+            ):
+                raise OverwriteNewerCloudError(
+                    f"File ({destination}) is newer than ({self}). "
+                    f"To overwrite "
+                    f"pass `force_overwrite_to_cloud=True`."
+                )
+
+            return self.client._move_file(self, destination, remove_src=False)
+
+        else:
+            if not destination.exists() or destination.is_file():
+                return destination.upload_from(
+                    self.fspath, force_overwrite_to_cloud=force_overwrite_to_cloud
+                )
+            else:
+                return (destination / self.name).upload_from(
+                    self.fspath, force_overwrite_to_cloud=force_overwrite_to_cloud
+                )
+
+    def copytree(
+        self,
+        destination: Union[str, os.PathLike, "CloudPath"],
+        force_overwrite_to_cloud: bool = False,
+    ) -> Union[Path, "CloudPath"]:
+        """Copy self to a directory, if self is a directory."""
+        if not self.is_dir():
+            raise CloudPathNotADirectoryError(
+                f"Origin path {self} must be a directory. To copy a single file use the method copy."
+            )
+
+        # handle string version of cloud paths + local paths
+        if isinstance(destination, (str, os.PathLike)):
+            destination = anypath.to_anypath(destination)
+
+        if destination.exists() and destination.is_file():
+            raise CloudPathFileExistsError(
+                "Destination path {destination} of copytree must be a directory."
+            )
+
+        destination.mkdir(parents=True, exist_ok=True)
+
+        for subpath in self.iterdir():
+            if subpath.is_file():
+                subpath.copy(
+                    destination / subpath.name, force_overwrite_to_cloud=force_overwrite_to_cloud
+                )
+            elif subpath.is_dir():
+                subpath.copytree(
+                    destination / subpath.name, force_overwrite_to_cloud=force_overwrite_to_cloud
+                )
+
+        return destination
+
     # ===========  private cloud methods ===============
     @property
     def _local(self):
@@ -620,7 +742,7 @@ class CloudPath(metaclass=CloudPathMeta):
     def _refresh_cache(self, force_overwrite_from_cloud=False):
         try:
             stats = self.stat()
-        except NoStat:
+        except NoStatError:
             # nothing to cache if the file does not exist; happens when creating
             # new files that will be uploaded
             return
@@ -639,7 +761,7 @@ class CloudPath(metaclass=CloudPathMeta):
             os.utime(self._local, times=(stats.st_mtime, stats.st_mtime))
 
         if self._dirty:
-            raise OverwriteDirtyFile(
+            raise OverwriteDirtyFileError(
                 f"Local file ({self._local}) for cloud path ({self}) has been changed by your code, but "
                 f"is being requested for download from cloud. Either (1) push your changes to the cloud, "
                 f"(2) remove the local file, or (3) pass `force_overwrite_from_cloud=True` to "
@@ -649,7 +771,7 @@ class CloudPath(metaclass=CloudPathMeta):
         # if local newer but not dirty, it was updated
         # by a separate process; do not overwrite unless forced to
         if self._local.stat().st_mtime > stats.st_mtime:
-            raise OverwriteNewerLocal(
+            raise OverwriteNewerLocalError(
                 f"Local file ({self._local}) for cloud path ({self}) is newer on disk, but "
                 f"is being requested for download from cloud. Either (1) push your changes to the cloud, "
                 f"(2) remove the local file, or (3) pass `force_overwrite_from_cloud=True` to "
@@ -657,45 +779,69 @@ class CloudPath(metaclass=CloudPathMeta):
             )
 
     def _upload_local_to_cloud(self, force_overwrite_to_cloud: bool = False):
+        """Uploads cache file at self._local to the cloud"""
         # We should never try to be syncing entire directories; we should only
         # cache and upload individual files.
         if self._local.is_dir():
             raise ValueError("Only individual files can be uploaded to the cloud")
 
+        uploaded = self._upload_file_to_cloud(
+            self._local, force_overwrite_to_cloud=force_overwrite_to_cloud
+        )
+
+        # force cache time to match cloud times
+        stats = self.stat()
+        os.utime(self._local, times=(stats.st_mtime, stats.st_mtime))
+
+        # reset dirty and handle now that this is uploaded
+        self._dirty = False
+        self._handle = None
+
+        return uploaded
+
+    def _upload_file_to_cloud(self, local_path, force_overwrite_to_cloud: bool = False):
+        """Uploads file at `local_path` to the cloud if there is not a newer file
+        already there.
+        """
         try:
             stats = self.stat()
-        except NoStat:
+        except NoStatError:
             stats = None
 
         # if cloud does not exist or local is newer or we are overwriting, do the upload
         if (
             not stats  # cloud does not exist
-            or (self._local.stat().st_mtime > stats.st_mtime)
+            or (local_path.stat().st_mtime > stats.st_mtime)
             or force_overwrite_to_cloud
         ):
             self.client._upload_file(
-                self._local,
+                local_path,
                 self,
             )
 
-            # force cache time to match cloud times
-            stats = self.stat()
-            os.utime(self._local, times=(stats.st_mtime, stats.st_mtime))
-
-            # reset dirty and handle now that this is uploaded
-            self._dirty = False
-            self._handle = None
-
             return self
 
         # cloud is newer and we are not overwriting
-        raise OverwriteNewerCloud(
+        raise OverwriteNewerCloudError(
             f"Local file ({self._local}) for cloud path ({self}) is newer in the cloud disk, but "
             f"is being requested to be uploaded to the cloud. Either (1) redownload changes from the cloud or "
             f"(2) pass `force_overwrite_to_cloud=True` to "
             f"overwrite."
         )
 
+    # ===========  pydantic integration special methods ===============
+    @classmethod
+    def __get_validators__(cls):
+        """Pydantic special method. See
+        https://pydantic-docs.helpmanual.io/usage/types/#custom-data-types"""
+        yield cls._validate
+
+    @classmethod
+    def _validate(cls, value: Any):
+        """Used as a Pydantic validator. See
+        https://pydantic-docs.helpmanual.io/usage/types/#custom-data-types"""
+        return cls(value)
+
 
 # The function resolve is not available on Pure paths because it removes relative
 # paths and symlinks. We _just_ want the relative path resolution for
